// vim: syn=c
/*
 * Minimalistic 1D Pong
 *
 * Copyright (C) 2018  B.Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/cpufunc.h>

#define DEBOUNCE_TIME		10	// Milliseconds to hold off bouncing
#define TIMEOUT_START		100	// milliseconds between start animations
#define TIMEOUT_BLINK		250	// Blink timeout (2 Hz)
#define TIMEOUT_MOVE		500	// Initial time between moves
#define TIMEOUT_MOVE_MIN	50	// Fastest move
#define TIMEOUT_MOVE_STEP	10	// Incremental speed-up
#define TIMEOUT_WIN		50	// Win animation speed
#define TIMEOUT_IDLE		7500	// Go to sleep after this period

#define TTIME_INIT		100	// 100 ms initial hit sound
#define TTIME_MOVE		100	// 100 ms move sound
#define TTIME_BOUNCE		100	// 100 ms bounce sound
#define TTIME_WIN		100	// 100 ms win sound

#define EV_BUTTON_L_PRESSED	0x01	// Bit-masks for button events
#define EV_BUTTON_R_PRESSED	0x02
#define EV_TIMER		0x04	// Timer event bit-mask
#define EV_TONETIMER		0x08
#define EV_IDLETIMER		0x10

// Timer0 TOP values for particular note frequencies @ 4 MHz
#define NOTE_F3		(179-1)	// ~174.6 Hz (-0.019%)
#define NOTE_G3		(159-1)	// ~196.0 Hz (+0.276%)
#define NOTE_G4		(80-1)	// ~392.0 Hz (-0.351%)
#define NOTE_C5		(60-1)	// ~520.8 Hz (-0.462%)

// Game states
enum {
	GS_IDLE,
	GS_START,
	GS_INIT_L,	GS_MOVE_LR,	GS_BOUNCE_R,	GS_WIN_L,
	GS_INIT_R,	GS_MOVE_RL,	GS_BOUNCE_L,	GS_WIN_R
};

#define NLEDS			6
#define LED_1			0	// Define LED index position
#define LED_2			1
#define LED_3			2
#define LED_4			3
#define LED_5			4
#define LED_6			5
#define LED_LEFT		LED_1
#define LED_RIGHT		LED_5
#define LED_GREEN		LED_6

static uint8_t leds[NLEDS];	// The leds -> 0x00 is off, 0xff if on (due to use of binary-& in output routine)

static uint16_t timer;		// Timer counter variable
static uint16_t tonetimer;	// Timer counter for tones
static uint16_t idletimer;	// Timer counter for going idle
static uint16_t timeoutmove;	// Speed tracking

static uint8_t animstate;	// Animation state counter
static uint8_t gamestate;	// Current state of the game
static uint8_t ledpos;		// Current LED position in move

static uint8_t oldtime;		// Previous timestamp

static volatile uint8_t msecs;	// millisecond counter
static volatile uint8_t adcval;	// Button state value from ADC


/*
 * Timer setup and shutdown control
 * Taken out separately to have better control over power consumption.
 */
// Timer0 is sound generator
static inline void timer0Setup()
{
	PRR &= ~_BV(PRTIM0);			// Enable timer peripheral clock
	TCCR0A = _BV(WGM01);			// CTC, OC0A disconnect
	TCCR0B = _BV(CS00) | _BV(CS01);		// 1:64
	OCR0A = 1;				// Just a value
	TCNT0 = 0;
}

static inline void timer0Shutdown()
{
	TCCR0B = 0;
	TCCR0A = 0;
	PRR |= _BV(PRTIM0);			// Disable timer peripheral clock
}

// Timer1 is millisecond timer
static inline void timer1Setup()
{
	PRR &= ~_BV(PRTIM1);			// Enable timer peripheral clock
	OCR1C = F_CPU / (64ul*1000) - 1;	// TOP; ~1 ms interval (about 1008 Hz @ 4 MHz source clock)
	OCR1A = 1;				// Timer interrupt early on
	TCCR1 = _BV(CTC1) | _BV(CS12) | _BV(CS11) | _BV(CS10);	// CTC mode 1:64 (62.5 kHz @ 4 MHz)
	GTCCR = 0;
	TCNT1 = 0;
	TIMSK |= _BV(OCIE1A);
}

static inline void timer1Shutdown()
{
	TIMSK &= ~_BV(OCIE1A);
	TCCR1 = 0;
	PRR |= _BV(PRTIM1);			// Disable timer peripheral clock
}

/*
 * Set individual LED state
 */
static inline void setLed(uint8_t idx, uint8_t val)
{
	leds[idx] = val ? 0xff : 0x00;
}

/*
 * Get current LED state
 */
static inline uint8_t ledState(uint8_t idx)
{
	return leds[idx] != 0;
}

/*
 * Turn all LEDs off
 */
static void ledsClear(void)
{
	setLed(LED_1, 0);
	setLed(LED_2, 0);
	setLed(LED_3, 0);
	setLed(LED_4, 0);
	setLed(LED_5, 0);
	setLed(LED_6, 0);
}

/*
 * Make music (well, noise)
 */
static inline void tone(uint8_t tmr, uint16_t len)
{
	tonetimer = len;
	OCR0A = tmr;
	TCNT0 = 0;
	TCCR0A = (TCCR0A & ~(_BV(COM0A1) | _BV(COM0A0))) | _BV(COM0A0);	// Toggle OC0A on match
}

static inline void noTone()
{
	TCCR0A = (TCCR0A & ~(_BV(COM0A1) | _BV(COM0A0))) | _BV(COM0A1);	// Clear OC0A on match
}

/*
 * The current time
 */
static inline uint8_t millis(void)
{
	return msecs;
}

/*
 * Animation in start state
 */
static void animateStart(void)
{
	// Pattern:
	//  0: x x o x x
	//  1: x o o o x
	//  2: o o o o o
	//  3: o o x o o
	//  4: o x x x o
	//  5: x x x x x
	//  6: o x x x o
	//  7: o o x o o
	//  8: o o o o o
	//  9: x o o o x
	// 10: x x o x x
	// 11: x x x x x
	switch(animstate) {
	case 0:
		setLed(LED_3, 1);
		break;
	case 1:
		setLed(LED_2, 1);
		setLed(LED_4, 1);
		break;
	case 2:
		setLed(LED_1, 1);
		setLed(LED_5, 1);
		break;
	case 3:
		setLed(LED_3, 0);
		break;
	case 4:
		setLed(LED_2, 0);
		setLed(LED_4, 0);
		break;
	case 5:
		setLed(LED_1, 0);
		setLed(LED_5, 0);
		break;
	case 6:
		setLed(LED_1, 1);
		setLed(LED_5, 1);
		break;
	case 7:
		setLed(LED_2, 1);
		setLed(LED_4, 1);
		break;
	case 8:
		setLed(LED_3, 1);
		break;
	case 9:
		setLed(LED_1, 0);
		setLed(LED_5, 0);
		break;
	case 10:
		setLed(LED_2, 0);
		setLed(LED_4, 0);
		break;
	case 11:
		setLed(LED_3, 0);
		/* Fallthrough */
	default:
		setLed(LED_6, !ledState(LED_6));	// Blink green
		animstate = -1;
		break;
	}
	animstate++;
}

/*
 * Animation in win state
 */
static void animateWin(uint8_t x)
{
	if(animstate >= 20)
		animstate = 0;

	if(animstate & 2)
		setLed(LED_GREEN, !ledState(LED_GREEN));

	if(x == GS_WIN_L) {
		if(animstate < 5)
			setLed(LED_LEFT + animstate, 1);
		else if(animstate < 10)
			setLed(LED_LEFT + (animstate - 5), 1);
		else if(animstate < 15)
			setLed(LED_LEFT + (animstate - 10), 0);
		else if(animstate < 20)
			setLed(LED_LEFT + (animstate - 15), 0);
	} else {
		if(animstate < 5)
			setLed(LED_RIGHT - animstate, 1);
		else if(animstate < 10)
			setLed(LED_RIGHT - (animstate - 5), 1);
		else if(animstate < 15)
			setLed(LED_RIGHT - (animstate - 10), 0);
		else if(animstate < 20)
			setLed(LED_RIGHT - (animstate - 15), 0);
	}
	animstate++;
}

/*
 * Perform timer countdown and return event if run down
 */
static inline uint8_t doTimer(uint8_t tdiff, uint16_t *tmr, uint8_t ev)
{
	if(0 != *tmr) {
		if(*tmr > tdiff) {
			// Timer is larger than time difference; therefore, still counting
			*tmr -= tdiff;	// Timer countdown
			return 0;
		} else {
			*tmr = 0;
			// Done counting, return event
			return ev;
		}
	}
	return 0;
}

//
// The buttons use a voltage divider.
//
// Adc level:
// --------+ ~255 No button pressed
//         |
//         |
//         |
//         +-------+ (~161) Button L pressed
//                 |
//                 +-------+ (~127) Button R pressed
//                         |
//                         +--------+ (~99) Both R and L buttons pressed
//                                  |
//

#define BST_N		0
#define BST_R		EV_BUTTON_R_PRESSED
#define BST_L		EV_BUTTON_L_PRESSED
#define BST_X		(BST_L | BST_R)

#define BUT_BOX		8					// +/- Match range
#define BUT_N_VAL	(240)					// No button; should be 255, but we want larger margin
#define BUT_L_VAL	(255 * 47000ul / (27000ul + 47000ul))	// Left pressed
#define BUT_R_VAL	(255 * 27000ul / (27000ul + 27000ul))	// Right pressed
#define BUT_X_VAL	(255 * 17149ul / (27000ul + 17149ul))	// Both pressed

static inline uint8_t doButtons(uint8_t tdiff)
{
	static uint8_t debouncetimer;	// Debounce timer
	static uint8_t bstate = BST_N;	// Current state
	uint8_t adc = adcval;		// Get the volatile value into a stable variable
	uint8_t nstate = bstate;	// New state

	if(!debouncetimer) {
		if(bstate != BST_N && adc >= BUT_N_VAL - BUT_BOX)
			nstate = BST_N;
		if(bstate != BST_R && adc >= BUT_R_VAL - BUT_BOX && adc <= BUT_R_VAL + BUT_BOX)
			nstate = BST_R;
		if(bstate != BST_L && adc >= BUT_L_VAL - BUT_BOX && adc <= BUT_L_VAL + BUT_BOX)
			nstate = BST_L;
		if(bstate != BST_X && adc >= BUT_X_VAL - BUT_BOX && adc <= BUT_X_VAL + BUT_BOX)
			nstate = BST_X;

		if(nstate != bstate) {
			debouncetimer = DEBOUNCE_TIME;
			uint8_t rv = (bstate ^ nstate) & nstate;
			bstate = nstate;
			return rv;
		}
	} else {
		if(tdiff >= debouncetimer)
			debouncetimer = 0;
		else
			debouncetimer -= tdiff;
	}
	return 0;
}

/*
 * Set the CPU into sleep mode (power down). This will reduce power consumption
 * and we'll awake by the watchdog timer in about 32 ms.
 */
static void doSleep(void)
{
	// Wait for a pending ADC conversion to finish
	loop_until_bit_is_clear(ADCSRA, ADSC);

	// Disable ADC interrupts
	// This prevents pending ADC conversion to wake us prematurely
	// The nops are to ensure that a pending interrupt is serviced
	_NOP();
	_NOP();
	_NOP();
	_NOP();
	ADCSRA &= ~_BV(ADIE);

	timer0Shutdown();	// Save more power, we need no timers here
	timer1Shutdown();
	PORTB = 0;		// Ensure that the LEDs are off if the timer has had no time to go round

	// Go power down
	// and keep doing so until a button is pressed
	do {
		wdt_reset();		// Get woken in due time
		sleep_enable();
		sleep_cpu();		// Zzzzzzz
		sleep_disable();
		wdt_reset();

		ADCSRA |= _BV(ADSC);	// Perform conversion to see if a button is pressed
		loop_until_bit_is_clear(ADCSRA, ADSC);
		adcval = ADCH;

	} while(adcval >= BUT_N_VAL - BUT_BOX);

	timer0Setup();		// Re-enable timers
	timer1Setup();
	ADCSRA |= _BV(ADIE);	// Re-enable ADC interrupt
}

/*
 * Perform game state switching
 * - Handle state exit actions
 * - Switch state
 * - Handle state entry actions
 */
static void stateSet(uint8_t newstate)
{
	// Exit actions
	switch(gamestate) {
	case GS_INIT_L:
	case GS_INIT_R:
		timeoutmove = TIMEOUT_MOVE;	// Slow speed start
		break;

	case GS_BOUNCE_L:
	case GS_BOUNCE_R:
		// Update speed
		if(timeoutmove > TIMEOUT_MOVE_MIN)
			timeoutmove -= TIMEOUT_MOVE_STEP;
		timer = timeoutmove;		// Setup move timer
		tone(NOTE_G3, TTIME_BOUNCE);
		setLed(LED_GREEN, 0);
		break;
	}

	// Switch to new state
	gamestate = newstate;

	// Entry actions
	switch(gamestate) {
	case GS_IDLE:
		ledsClear();		// All LEDs off
		noTone();
		break;

	case GS_START:
		ledsClear();		// All LEDs off
		timer = TIMEOUT_START;
		idletimer = TIMEOUT_IDLE;
		animstate = 0;
		break;

	case GS_INIT_L:
	case GS_INIT_R:
		timer = TIMEOUT_BLINK;	// Setup blink timer
		ledsClear();		// All LEDs off
		setLed(gamestate == GS_INIT_L ? LED_LEFT : LED_RIGHT, 1);
		break;

	case GS_MOVE_LR:
		setLed(LED_LEFT, 1);	// Left LED on
		ledpos = LED_LEFT;	// Setup LED position
		timer = timeoutmove;	// Setup move timer
		tone(NOTE_F3, TTIME_INIT);	// Serve tone
		break;

	case GS_MOVE_RL:
		setLed(LED_RIGHT, 1);	// Right LED on
		ledpos = LED_RIGHT;	// Setup LED position
		timer = timeoutmove;	// Setup move timer
		tone(NOTE_F3, TTIME_INIT);	// Serve tone
		break;

	case GS_BOUNCE_L:
	case GS_BOUNCE_R:
		setLed(LED_GREEN, 1);	// Indicate hit-back enable
		break;

	case GS_WIN_L:
	case GS_WIN_R:
		tone(NOTE_C5, TTIME_WIN);	// Win tone
		animstate = 0;
		timer = TIMEOUT_WIN;	// Winning animation timer
		break;
	}
}

/*
 * - init variables
 */
static void setup(void)
{
	oldtime = millis();		// Initialize timestamp
	doButtons(0);			// Initial button state
	stateSet(gamestate = GS_START);	// Initial game state
}

/*
 * Main loop
 * - debounce buttons
 * - run timer
 * - run state machine
 */
static void loop(void)
{
	uint8_t now = millis();		// Current time in milliseconds
	uint8_t tdiff = now - oldtime;	// Time difference from previous loop
	uint8_t events = 0;		// Events seen in this pass
	uint8_t oldstate = gamestate;	// See if we change states (for idle detection)

	wdt_reset();

	if(tdiff > 0) {
		// At least one millisecond has passed
		oldtime = now;

		// Collect the (async) button events
		events |= doButtons(tdiff);

		// Timer counter handling
		events |= doTimer(tdiff, &timer, EV_TIMER);
		events |= doTimer(tdiff, &tonetimer, EV_TONETIMER);
		events |= doTimer(tdiff, &idletimer, EV_IDLETIMER);

		ADCSRA |= _BV(ADSC);	// Start next AD conversion to do button input
	}

	// Sound off handling is required in all states, just as well do it here.
	if(events & EV_TONETIMER)
		noTone();

	// Run the state machine logic
	switch(gamestate) {
	case GS_IDLE:
		if(events & (EV_BUTTON_L_PRESSED | EV_BUTTON_R_PRESSED)) {
			stateSet(GS_START);
		} else {
			doSleep();
		}
		break;
	case GS_START:
		if(events & EV_TIMER) {
			animateStart();			// Do some animation
			timer = TIMEOUT_START;		// Setup timer again
		}
		if(events & (EV_BUTTON_L_PRESSED | EV_BUTTON_R_PRESSED)) {
			// Either button pressed
			stateSet((events & EV_BUTTON_L_PRESSED) ? GS_INIT_L : GS_INIT_R);
		}
		else if(events & EV_IDLETIMER) {
			// If the idle timer hits, then we need to sleep
			stateSet(oldstate = GS_IDLE);	// Force idle state
		}
		break;

	case GS_INIT_L:
		if(events & EV_TIMER) {
			setLed(LED_LEFT, !ledState(LED_LEFT));	// Invert
			timer = TIMEOUT_BLINK;		// Setup timer again
		}
		if(events & EV_BUTTON_L_PRESSED) {
			stateSet(GS_MOVE_LR);		// Next state, move left to right
		} else if(events & EV_IDLETIMER) {
			stateSet(GS_START);
		}
		break;

	case GS_MOVE_LR:
		if(events & EV_TIMER) {
			setLed(ledpos, 0);		// Previous off
			setLed(++ledpos, 1);		// Next on
			tone(NOTE_G4, TTIME_MOVE);	// Move tone
			if(ledpos == LED_RIGHT) {
				stateSet(GS_BOUNCE_R);	// Reached end of line
			}
			timer = timeoutmove;		// Setup timer either move or bounce
		}
		break;

	case GS_BOUNCE_R:
		if(events & EV_TIMER) {
			// Timeout, left wins
			stateSet(GS_WIN_L);		// Set winning state
		} else if(events & EV_BUTTON_R_PRESSED) {
			stateSet(GS_MOVE_RL);		// Next state, move right to left
		}
		break;

	case GS_WIN_L:
		if(events & EV_TIMER) {
			animateWin(GS_WIN_L);
			timer = TIMEOUT_WIN;		// Winning animation timer
		}
		if(events & EV_IDLETIMER) {
			stateSet(GS_START);
		} else if(events & EV_BUTTON_L_PRESSED) {
			stateSet(GS_INIT_L);		// Next state, move left to right
		}
		break;

	case GS_INIT_R:
		if(events & EV_TIMER) {
			setLed(LED_RIGHT, !ledState(LED_RIGHT));	// Invert
			timer = TIMEOUT_BLINK;		// Setup timer again
		}
		if(events & EV_BUTTON_R_PRESSED) {
			stateSet(GS_MOVE_RL);		// Next state, move right to left
		} else if(events & EV_IDLETIMER) {
			stateSet(GS_START);
		}
		break;

	case GS_MOVE_RL:
		if(events & EV_TIMER) {
			setLed(ledpos, 0);		// Previous off
			setLed(--ledpos, 1);		// Next on
			tone(NOTE_G4, TTIME_MOVE);	// Move tone
			if(ledpos == LED_LEFT) {
				stateSet(GS_BOUNCE_L);	// Reached end of line
			}
			timer = timeoutmove;		// Setup timer either move or bounce
		}
		break;

	case GS_BOUNCE_L:
		if(events & EV_TIMER) {
			// Timeout, right wins
			stateSet(GS_WIN_R);		// Set winning state
		} else if(events & EV_BUTTON_L_PRESSED) {
			stateSet(GS_MOVE_LR);		// Next state, move left to right
		}
		break;

	case GS_WIN_R:
		if(events & EV_TIMER) {
			animateWin(GS_WIN_R);
			timer = TIMEOUT_WIN;		// Winning animation timer
		}
		if(events & EV_IDLETIMER) {
			stateSet(GS_START);
		} else if(events & EV_BUTTON_R_PRESSED) {
			stateSet(GS_INIT_R);		// Next state, move right to left
		}
		break;

	default:
		// Should not get here, but just in case we reset the state
		stateSet(GS_START);
		break;
	}

	// Reset the idle timer if something is happening
	if(gamestate != oldstate)
		idletimer = TIMEOUT_IDLE;
}

// The DDR register setting for the LEDs
static const uint8_t charliedir[NLEDS] PROGMEM = {
	_BV(PB1) | _BV(PB2),
	_BV(PB1) | _BV(PB2),
	_BV(PB2) | _BV(PB3),
	_BV(PB2) | _BV(PB3),
	_BV(PB3) | _BV(PB1),
	_BV(PB3) | _BV(PB1),
};

// The PORT high/low settings for the LEDs
static const uint8_t charlieval[NLEDS] PROGMEM = {
	_BV(PB1),
	_BV(PB2),
	_BV(PB2),
	_BV(PB3),
	_BV(PB3),
	_BV(PB1),
};

/*
 * Timer 1 OC1A match interrupt
 * Should fire every millisecond or so.
 */
ISR(TIM1_COMPA_vect)
{
	static uint8_t charlieidx;

	msecs++;	// Update time

	//
	// Update charlieplex leds
	// - turn them all off
	// - set the DDR to address the next led
	// - set the PORT to the enable the next led or zero if it should be off
	// The sequence of statements forces the compiler to cache the values
	// before operating on the DDR and PORT registers because the sequence
	// of statements is retained. This makes the switch all the faster and
	// we lose minimal time in the switch of charlieplex (11 clock-cycles).
	//
	if(++charlieidx >= NLEDS) {
		charlieidx = 0;
	}
#define LEDBITS	(_BV(PB1) | _BV(PB2) | _BV(PB3))
#define LEDMASK	(~LEDBITS)
	uint8_t ddr = pgm_read_byte(&charliedir[charlieidx]);
	uint8_t prt = pgm_read_byte(&charlieval[charlieidx]);
	uint8_t led = leds[charlieidx];
	PORTB &= LEDMASK;			// Turn all off
	DDRB = ddr | (DDRB & LEDMASK);		// Change directions
	PORTB |= prt & led;			// Turn on next
}

/*
 * ADC conversion done interrupt
 */
ISR(ADC_vect)
{
	adcval = ADCH;
}

/*
 * Watchdog will wake us from sleep
 */
ISR(WDT_vect)
{
	// The doc specifies that we need to re-enable the interrupt enable flag
	WDTCR |= _BV(WDIE);	// Re-enable use of interrupt
}

/*
 * Disable the WDT.
 * This will prevent an endless reset loop if the WDT is enable after a reset.
 * The code is executed very early on after boot.
 *
 * See avr-libc documentation (avr/wdt.h) for details on this routine.
 */
void get_mcusr(void) __attribute__((naked)) __attribute__((section(".init3")));
void get_mcusr(void)
{
	MCUSR = 0;
	wdt_disable();
}

/*
 * Setup MCU
 */
static void init(void)
{
	cli();	// Just in case

	CLKPR = _BV(CLKPCE);	// Enable clock update
	CLKPR = _BV(CLKPS0);	// 1:2 clock -> 4 MHz

	sei();

	PRR = _BV(PRUSI);	// Shut down USI module

	PORTB = 0;
	DDRB = 0b00000001;	// Buzzer output

	DIDR0 = _BV(ADC2D);	// Disable digital input on button input

	ACSR = 0;				// Disable Analog Comparator
	ADMUX = _BV(MUX1) | _BV(ADLAR);		// ADC2 channel on PB4, Vcc reference, left adjust
	ADCSRA = _BV(ADPS1) | _BV(ADPS0);	// 1:8 prescale (500 kHz is fine for 8 bit)
	ADCSRB = 0;
	ADCSRA |= _BV(ADEN);			// Enable ADC

	ADCSRA |= _BV(ADSC);			// Perform first conversion
	loop_until_bit_is_clear(ADCSRA, ADSC);
	adcval = ADCH;				// Discard the result

	ADCSRA |= _BV(ADSC);			// Perform second conversion (initial state for buttons)
	loop_until_bit_is_clear(ADCSRA, ADSC);
	adcval = ADCH;				// Save as initial button value
	ADCSRA |= _BV(ADIE);			// Enable ADC interrupts

	timer0Setup();		// Setup timers
	timer1Setup();
	msecs = 0;

	wdt_enable(WDTO_30MS);	// 32 ms interval
	WDTCR |= _BV(WDIE);	// Use interrupt and not reset

	set_sleep_mode(SLEEP_MODE_PWR_DOWN);
}

int main(void)
{
	init();
	setup();
	while(1) {
		loop();
	}
}
